package edg.edge;

import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import edg.graph.EDG;
import edg.graph.Node;
import edg.graph.Variable;

public abstract class EdgeGenerator {
//	public static void generateEdges(EDG edg)
//	{
//		new ControlFlowEdgeGenerator(edg).generate();
//		new ControlEdgeGenerator(edg).generate();
//		new InterproceduralEdgeGenerator(edg).generate();
//		new FlowEdgeGenerator(edg).generate();
//		new ValueEdgeGenerator(edg).generateJava();
//		new SummaryEdgeGenerator(edg).generate();
//		// EXCEPTIONS
//		new ExceptionEdgeGenerator(edg).generate();
//	}

	protected final EDG edg;

	public EdgeGenerator(EDG edg)
	{
		this.edg = edg;
	}

	public abstract void generate();

	protected List<Node> getImplicitRestrictions(List<Node> patterns)
	{
		final List<Node> implicitRestrictions = new LinkedList<>();

		for (Node pattern : patterns)
		{
			final List<Node> patternRestrictions = this.getImplicitRestrictions(pattern);

			implicitRestrictions.addAll(patternRestrictions);
		}

		return implicitRestrictions;
	}
	protected List<Node> getImplicitRestrictions(Node pattern)
	{
		final List<Node> implicitRestrictions = new LinkedList<>();

		pattern = pattern.getType() == Node.Type.Expression ? edg.getChild(pattern, 0) : pattern;

		switch (pattern.getType())
		{
			case Literal:
				implicitRestrictions.add(pattern);
				break;
			case Variable:
				final Variable variable = (Variable) pattern;
				if (variable.getContext() == Variable.Context.Use)
					implicitRestrictions.add(pattern);
				break;
			case List:
			case DataConstructor:
				final List<Node> children = edg.getChildrenNonResult(pattern);
				implicitRestrictions.add(pattern);
				for (Node child : children)
					implicitRestrictions.addAll(this.getImplicitRestrictions(child));
				break;
			default:
				throw new RuntimeException("Node type not contemplated: " + pattern.getType());
		}

		return implicitRestrictions;
	}
	protected List<Node[]> getMatches(Node pattern, Node expression)
	{
		final List<Node[]> matches = new LinkedList<>();

		final Node.Type patternType = pattern.getType();
		final Node.Type expressionType = expression.getType();
		final String patternText = pattern.getName();
		final String expressionText = expression.getName();

		if (expressionType == Node.Type.Variable ||
				((patternType == Node.Type.Literal) &&
						(expressionType == Node.Type.Literal) &&
						patternText.equals(expressionText)))
			matches.add(new Node[] { pattern, expression } );
		else		if (patternType == Node.Type.Variable ||
				((patternType == Node.Type.Literal) && (expressionType == Node.Type.Operation || expressionType == Node.Type.Call)) ||
				(patternType == Node.Type.DataConstructor && expressionType == Node.Type.Call) ||
				(patternType == Node.Type.List && (expressionType == Node.Type.Operation || expressionType == Node.Type.Call || expressionType == Node.Type.ListComprehension)))
		{
			final Node result = edg.getResFromNode(expression);
			if (result != null)
				matches.add(new Node[] { pattern, result } );
		}
		else		if ((patternType == Node.Type.Literal || patternType == Node.Type.DataConstructor || patternType == Node.Type.List) &&
				(expressionType == Node.Type.Switch || expressionType == Node.Type.If || expressionType == Node.Type.Equality || expressionType == Node.Type.Block))
		{
			final Node resultRoot = edg.getResFromNode(expression);
			if (resultRoot != null)
				matches.addAll(this.getMatches(pattern, resultRoot));
		}
		else if (patternType == Node.Type.Equality)
		{
			final Node result = edg.getResFromNode(pattern);
			if (result != null)
				matches.addAll(this.getMatches(result, expression));
		}
		else if ((patternType == Node.Type.DataConstructor || patternType == Node.Type.List) && expressionType == patternType)
		{
			final List<Node> patternNodeChildren = edg.getChildrenNonResult(pattern);
			final List<Node> expressionNodeChildren = edg.getChildrenNonResult(expression);

			if (patternNodeChildren.size() == expressionNodeChildren.size())
			{
				final List<Node[]> childrenMatches = new LinkedList<>();

				for (int childIndex = 0; childIndex < patternNodeChildren.size(); childIndex++)
				{
					final Node patternChild = patternNodeChildren.get(childIndex);
					final Node expressionChild = expressionNodeChildren.get(childIndex);

					childrenMatches.addAll(this.getMatches(patternChild, expressionChild));
				}

				if (!childrenMatches.isEmpty())
				{
					matches.add(new Node[] { pattern, expression } );
					matches.addAll(childrenMatches);
				}
			}
		}

		return matches;
	}

	protected MatchStatus isMatch(Node parameter, Node argument, Map<String,String> variableMap) {
		final Node.Type parameterType = parameter.getType();
		final String parameterText = parameter.getName();
		final Node.Type argumentType = argument.getType();
		final String argumentText = argument.getName();

		// Parameter = Variable (X) / Argument = Literal (1)
		if (parameterType == Node.Type.Variable && argumentType == Node.Type.Literal) {
			// First occurrence of the variable in the pattern
			if (((Variable) parameter).getContext() == Variable.Context.Definition) {
				if (!parameterText.equals("_"))
					variableMap.put(parameterText, argumentText);
				return new MatchStatus(true, true);
			} else { // The value associated to this pattern variable must be the same or there is no match
				assert variableMap.containsKey(parameterText);
				if (parameterText.equals("_"))
					return new MatchStatus(true, true);
				;
				String previousValue = variableMap.get(parameterText);
				return new MatchStatus(argumentText.equals(previousValue), true);
			}
		} else if (parameterType == Node.Type.Literal && argumentType == Node.Type.Literal &&
				parameterText.equals(argumentText)) { // Parameter = Literal (1) / Argument = Literal (1)
			return new MatchStatus(true, true);
		} else if (parameterType == Node.Type.Variable) { // Parameter = Variable / Argument = !Literal
			return new MatchStatus(true, !variableMap.containsKey(parameterText));
		} else if (// Parameter = Literal (1) / Argument = Operation || Call (A+B || foo())
				(parameterType == Node.Type.Literal && (argumentType == Node.Type.Operation || argumentType == Node.Type.Call)) ||
				// Parameter = Tuple ({A,B}) / Argument = Call (foo())
				(parameterType == Node.Type.DataConstructor && argumentType == Node.Type.Call) ||
				// Parameter = List ([A,B]) / Argument = Operation || Call || LC ([A] ++ [B] || foo() || [X || X <- [A,B]] )
				(parameterType == Node.Type.List && (argumentType == Node.Type.Operation || argumentType == Node.Type.Call || argumentType == Node.Type.ListComprehension))) {
			return new MatchStatus(true, false);
		} else if (parameterType == Node.Type.Equality) { // Pattern = Equality (X = Y) / Argument = ANY
			final Node pattern = edg.getChild(parameter, Node.Type.Pattern);
			final Node value = edg.getChild(parameter, Node.Type.Value);
			MatchStatus patternMatch = this.isMatch(pattern, argument, variableMap);
			MatchStatus valueMatch = this.isMatch(value, argument, variableMap);
			return new MatchStatus(patternMatch.isMatch() && valueMatch.isMatch(),
					patternMatch.isCompleteMatch() && valueMatch.isCompleteMatch());
		} else if ((parameterType == Node.Type.DataConstructor || // Parameter = Tuple ({X,Y}) || List ([X,Y])
				parameterType == Node.Type.List) && argumentType == parameterType) { // Argument = Same as pattern
			final List<Node> parameterChildren = edg.getChildrenNonResult(parameter);
			final List<Node> argumentChildren = edg.getChildrenNonResult(argument);

			if (parameterChildren.size() == argumentChildren.size())
			{
				boolean childrenMatches = true;
				boolean completeMatch = true;

				for (int childIndex = 0; childIndex < parameterChildren.size() && childrenMatches; childIndex++)
				{
					final Node parameterChild = parameterChildren.get(childIndex);
					final Node argumentChild = argumentChildren.get(childIndex);

					MatchStatus childMatch = isMatch(parameterChild, argumentChild, variableMap);

					childrenMatches = childrenMatches && childMatch.isMatch();
					completeMatch = completeMatch && childMatch.isCompleteMatch();
				}
				return new MatchStatus(childrenMatches,completeMatch);
			}
			return new MatchStatus();
		} else if ((parameterType == Node.Type.Literal || parameterType == Node.Type.DataConstructor || parameterType == Node.Type.List) && // Parameter = Literal || Tuple || List
				(argumentType == Node.Type.Switch || argumentType == Node.Type.Equality || argumentType == Node.Type.Block)) // Argument = Complex Expressions (Equality, Switch, Block)
		{
			boolean anyMatch = false;
			boolean allMatch = true;
			List<Node> valueExpressions = this.getLastExpression(argument);
			for (Node valueExpression : valueExpressions){
				MatchStatus matchValue = isMatch(parameter, valueExpression, variableMap);
				anyMatch = anyMatch || matchValue.isMatch();
				allMatch = allMatch && matchValue.isCompleteMatch();
			}
			return new MatchStatus(anyMatch, allMatch);
		} else if (argumentType == Node.Type.Variable) // Pattern = ANY / Argument = Variable (X) (ALWAYS MATCH)
			return new MatchStatus(true,false);
		return new MatchStatus();
	}

	private List<Node> getLastExpression(Node node){
		switch (node.getType()){
			case Switch:
				Node cases = edg.getChild(node, Node.Type.Cases);
				return getLastExpression(cases);
			case Cases:
				List<Node> result = new LinkedList<>();
				final List<Node> caseList = edg.getChildren(node);
				for (Node _case : caseList)
					result.addAll(getLastExpression(_case));
				return result;
			case Case:
				final Node body = edg.getChild(node, Node.Type.Body);
				return getLastExpression(body);
			case Body:
			case Block:
				List<Node> returnNodes = edg.getChildren(node, Node.Type.Return);
				List<Node> expressionNodes =
						returnNodes.stream().map(n -> edg.getChild(n, Node.Type.Value))
						.collect(Collectors.toList());
				assert expressionNodes.size() == 1;
				return expressionNodes;
			case Equality:
				return List.of(edg.getChild(node, Node.Type.Value));
			default:
				return List.of();
		}
	}

	protected List<Node> getVariables(String id, Variable.Context context, Boolean global)
	{
		final List<Node> variableNodes = new LinkedList<>();
		final List<Node> variables = this.edg.getNodes(node -> node instanceof Variable);

		for (Node variable : variables)
		{
			final Variable info = (Variable) variable;
			final String variableId = this.getId(variable);
			if (id != null && !variableId.equals(id))
				continue;
			if (context != null && info.getContext() != context)
				continue;
			if (global != null && info.isGlobal() != global)
				continue;
			variableNodes.add(variable);
		}

		return variableNodes;
	}

	protected List<Node> getVariables(String id, Variable.Context context, String className, Boolean global)
	{
		final List<Node> variableNodes = new LinkedList<>();
		final List<Node> variables = this.edg.getNodes(node -> node instanceof Variable);

		for (Node variable : variables)
		{
			final Variable info = (Variable) variable;
			final String variableId = this.getId(variable);
			if (id != null && !variableId.equals(id))
				continue;
			if (context != null && context == Variable.Context.Declaration && !info.isDeclaration())
				continue;
			else if (context != null && context != Variable.Context.Declaration && info.getContext() != context)
				continue;
			if (global != null && info.isGlobal() != global)
				continue;
			if (className != null && !className.equals(info.getInfo().getClassName()))
				continue;
			variableNodes.add(variable);
		}

		return variableNodes;
	}

	protected String getId(Node variable)
	{
		if (!(variable instanceof Variable))
			throw new RuntimeException("The node is not a variable");

		final Variable variableInfo = (Variable) variable;
		return variableInfo.getName();
	}
	protected boolean sameVariables(String id1, String id2)
	{
		final int id1ArrayIndex = id1.indexOf('[');
		final int id2ArrayIndex = id2.indexOf('[');
		final String identifier1 = id1ArrayIndex == -1 ? id1 : id1.substring(0,  id1ArrayIndex);
		final String identifier2 = id2ArrayIndex == -1 ? id2 : id2.substring(0,  id2ArrayIndex);

		return identifier1.equals(identifier2);
	}

	protected class MatchStatus{
		private boolean match;
		private boolean complete;

		public MatchStatus(){
			this(false,false);
		}
		public MatchStatus(boolean match, boolean complete){
			this.match = match;
			this.complete = complete;
		}

		public boolean isMatch() { return match; }
		public boolean isCompleteMatch() { return complete; }

	}
}
