package edg.edge;

import java.util.*;

import edg.constraint.EmptyConstraint;
import edg.graph.*;
import edg.traverser.ControlFlowTraverser;

public class FlowEdgeGenerator extends EdgeGenerator
{
	/** Used in the CFG traversal to generate Flow dependencies.
	 * Contains all the info related to variable definitions */
	public static class EdgeGenState {
		private final Node node;
		private final Map <String,Node> definitionMap;
		private final Map <String,Node> shadowDefinitionMap;
		private EdgeGenState(Node node)
		{
			this(node, new HashMap<>(), new HashMap<>());
		}

		private EdgeGenState(Node node, Map<String, Node> definitions, Map<String,Node> shadowDefinitions)
		{
			this.node = node;
			this.definitionMap = definitions;
			this.shadowDefinitionMap = shadowDefinitions;
		}
		public String toString() {
			return "Node" + this.node.getId() + ", DefVars" + this.definitionMap.keySet();
		}
		public boolean equals(Object o)
		{
			if (o == this)
				return true;
			if (!(o instanceof EdgeGenState))
				return false;

			final EdgeGenState state = (EdgeGenState) o;

			return Objects.equals(this.node, state.node) &&
					Objects.equals(this.definitionMap, state.definitionMap);
		}
		public int hashCode() {
			return Objects.hash(node.getId(), definitionMap);
		}
	}


	public FlowEdgeGenerator(EDG edg) {
		super(edg);
	}

	// ---------------------------------------------------------------- //
	// ------ TRAVERSE THE CFG TO GENERATE FLOW DEPENDENCY EDGES ------ //
	// ---------------------------------------------------------------- //

	public void generate() {
		final List<Node> clauses = edg.getNodes(Node.Type.Clause);

		for (Node clause : clauses)
			this.generateFlowEdges(clause);
	}

	/** Traverse every method definition generating flow
	 * dependencies between definitions and uses */
	public void generateFlowEdges(Node startNode)
	{
		final Node lastNode = edg.getResFromNode(startNode);
		final Set<EdgeGenState> doneWorks = new HashSet<>();
		final Set<EdgeGenState> pendingWorks = new HashSet<>();
		pendingWorks.add(new EdgeGenState(startNode));

		while(!pendingWorks.isEmpty()) {
			final EdgeGenState work =  pendingWorks.iterator().next();
			final Node workNode = work.node;
			final Map<String, Node> workDefinitionMap = work.definitionMap;
			final Map<String, Node> workShadowDefinitionMap = work.shadowDefinitionMap;
			pendingWorks.remove(work);

			if (workNode == lastNode)
				continue;
			if (doneWorks.contains(work))
				continue;

			if (workNode.getType() == Node.Type.Variable) {
				final Variable v = (Variable) workNode;
				switch(v.getContext())
				{
					case Def_Use:
						throw new RuntimeException("Def_Use variables does not exist in Erlang");
					case Definition:
						if (v.isShadowed())
							workShadowDefinitionMap.put(v.getName(),v);
						else
							workDefinitionMap.put(v.getName(),v);
						break;
					case Use:
						final Node definition = workShadowDefinitionMap.containsKey(v.getName()) ?
								workShadowDefinitionMap.get(v.getName()) : workDefinitionMap.get(v.getName());
						if (definition != null)
							this.addFlowEdge(definition, v);
					default:
						break;
				}
			}
			if (workNode.getType() == Node.Type.ListComprehension) // SHADOW ENVIRONMENT EXIT NODES
				workShadowDefinitionMap.clear();
			doneWorks.add(work);
			final Set<Node> nextNodes = ControlFlowTraverser.step(this.edg, workNode, LAST.Direction.Forwards);
			nextNodes.forEach(nextNode -> pendingWorks.add(new EdgeGenState(nextNode, new HashMap<>(workDefinitionMap), new HashMap<>(workShadowDefinitionMap))));
		}
	}

	private void addFlowEdge(Node source, Variable target) {
		if (!(source instanceof Variable))
			throw new RuntimeException("The definition is not a variable");

		final Node sourceRes = edg.getResFromNode(source);
		final Node targetRes = edg.getResFromNode(target);
		addIfNotExists(sourceRes, targetRes, Edge.Type.Flow);
	}

	private void addIfNotExists(Node source, Node target, Edge.Type edgeType){
		final Set<Edge> edges = edg.getEdges(source, LAST.Direction.Forwards, edgeType);
		for (Edge e : edges)
			if (edg.getEdgeTarget(e) == target)
				return;
		edg.addEdge(source, target, edgeType, EmptyConstraint.getConstraint());
	}
}