package edg.edge;

import java.util.*;

import edg.LASTBuilder.ClassInfo;
import edg.constraint.GlobalVariableConstraint;
import edg.constraint.PhaseConstraint;
import edg.constraint.SeekingConstraint;
import edg.graph.EDG;
import edg.graph.Edge;
import edg.graph.LAST.Direction;
import edg.graph.Node;
import edg.slicing.Phase;

public class InterproceduralEdgeGenerator extends EdgeGenerator
{
	public InterproceduralEdgeGenerator(EDG edg)
	{
		super(edg);
	}

	public void generate()
	{
		final List<Node> calls = edg.getNodes(Node.Type.Call);

		for (Node call : calls)
		{
			this.generateInputEdges(call);
			this.generateOutputEdges(call);
		}
	}

	/************************************/
	/************ Input edges ***********/
	/************************************/
	private void generateInputEdges(Node call)
	{
		final List<Node> possibleClauses = this.getPossibleClauses(call);
		final List<Node> matchingClauses = this.getMatchingClauses(possibleClauses, call);
		final Node callee = edg.getChild(call, Node.Type.Callee);
		final Node calleeResultNode = edg.getResFromNode(callee);
		final Node arguments = edg.getChild(call, Node.Type.Arguments);
		final Node argumentsIn = edg.getChild(call, Node.Type.ArgumentIn);
		final List<Node> argumentNodes = edg.getChildren(arguments);

		for (Node matchingClause : matchingClauses)
		{
			final Node parameters = edg.getChild(matchingClause, Node.Type.Parameters);
			final List<Node> parameterNodes = edg.getChildren(parameters);

			this.edg.addEdge(calleeResultNode, matchingClause, new Edge(Edge.Type.Call, new PhaseConstraint(Phase.Input)));
			for (int argumentIndex = 0; argumentIndex < argumentNodes.size(); argumentIndex++)
			{
				final Node argument = argumentNodes.get(argumentIndex);
				final Node parameter = parameterNodes.get(argumentIndex);
				final Node argumentResult = edg.getResFromNode(argument);
				final Node parameterResult = edg.getResFromNode(parameter);

				if (argumentResult != null && parameterResult != null)
					this.edg.addEdge(argumentResult, parameterResult, new Edge(Edge.Type.Input, new PhaseConstraint(Phase.Input)));
			}
			final Node parameterIn = edg.getChild(matchingClause, Node.Type.ParameterIn);
			//this.edg.addEdge(arguments, parameters, 0, new Edge(Edge.Type.Call, new PhaseConstraint(Phase.Input)));
			this.edg.addEdge(argumentsIn, parameterIn, new Edge(Edge.Type.Call, new PhaseConstraint(Phase.Input)));
		}
	}
	private List<Node> getPossibleClauses(Node call)
	{
		final Node callee = edg.getChild(call, Node.Type.Callee);
		final Node scopeNode = edg.getChild(callee, Node.Type.Scope);
		final Node nameNode = edg.getChild(callee, Node.Type.Name);
		final List<Node> scopeChildren = edg.getChildren(scopeNode);
		final Node routineArguments = edg.getChild(call, Node.Type.Arguments);
		final List<Node> arguments = edg.getChildren(routineArguments);

		// Module
		final Node moduleRef0 = scopeChildren.isEmpty() ? edg.getAncestor(call, Node.Type.Module) : scopeChildren.get(0);
		final Node moduleRef1 = moduleRef0.getType() != Node.Type.Expression ? moduleRef0 : edg.getChild(moduleRef0, Node.Type.Value);
		final Node moduleRef = moduleRef1.getType() != Node.Type.TypeTransformation ? moduleRef1 : edg.getChild(edg.getChild(moduleRef1, Node.Type.Variable),Node.Type.Value);
		final Node.Type moduleRefType = moduleRef.getType();
		
		//String moduleName = moduleRefType == Node.Type.Literal ? moduleRef.getName() : null;//moduleRef.getInfo().getClassName();
		final String moduleName0 = moduleRefType == Node.Type.Variable ? moduleRef.getInfo().getInfo()[1].toString() : null;
		final String moduleName1 = moduleRefType == Node.Type.Literal ? moduleRef.getName() : moduleName0;
		String moduleName = scopeChildren.size() == 0 ? moduleRef.getName() : moduleName1;
		
		// Function
		final Node name0 = edg.getChild(nameNode, 0);
		final Node name = name0.getType() != Node.Type.Expression ? name0 : edg.getChild(name0, Node.Type.Value);
		final Node.Type nameType = name.getType();
		String routineName = name.getName();

		// - M:R()	all routines of all modules				=> _:_
		// - m:R()	all routines of module m				=> m:_
		// - M:r()	routine r of all modules				=> _:r
		// - m:r()	routine f of module m					=> m:r
		// - r()	routine f of current module				=> m:r
		// - ar()	this anonymous routine					=> null:_
		// - X()	all routines (including anonymous ones)	=> null:null
		if (moduleRefType != Node.Type.Variable && moduleRefType != Node.Type.Literal && moduleRefType != Node.Type.Module)
			moduleName = "_";
		if (nameType != Node.Type.Literal)
			routineName = "_";
		if (nameType != Node.Type.Literal && scopeChildren.isEmpty())
			moduleName = null;
		if (moduleName == null && nameType != Node.Type.Routine)
			routineName = null;
		
//		if (moduleName.equals("_"))
//		{
//			final Node callModule = edg.getAncestor(call, Node.Type.Module);
//			final ClassInfo moduleInfo = (ClassInfo) callModule.getInfo().getInfo()[2];
//			final List<Node> classClauses = this.getAllClauses(moduleInfo, routineName);
//			System.out.println("STOP");
//			return classClauses;
//		}
//		else 
		if (moduleName != null)
		{
			final String moduleScopeName = moduleName;
			if (moduleScopeName.equals("super"))
			{
				final Node module = edg.getAncestor(call, Node.Type.Module);
				moduleName = module.getName();
			}
			final List<Node> modules = edg.getNodes(Node.Type.Module);
			for (Node module : modules)
			{
				final String moduleText = module.getName();
				if (moduleName.equals(moduleText))
				{
					final ClassInfo moduleInfo = (ClassInfo) module.getInfo().getInfo()[2];
					final List<Node> classClauses;
					if (!scopeChildren.isEmpty())
					{
						if (moduleScopeName.equals("super") && routineName.equals("<constructor>"))
							routineName = moduleScopeName+routineName;
						classClauses = this.getAllClauses(moduleInfo, routineName);
					}
					else
					{	
						final Node parentRoutineNode = edg.getAncestor(call, Node.Type.Routine);
						if (parentRoutineNode != null)
						{
							final String parentRoutineName = edg.getAncestor(call, Node.Type.Routine).getName();
							classClauses = this.getClassClauses(moduleInfo, routineName, parentRoutineName);
						}
						else 
							classClauses = this.getAllClauses(moduleInfo, routineName);
					}
					return classClauses;
				}
			}
		}
		
		final List<Node> possibleClauses = new LinkedList<Node>();
		final List<Node> clauses = edg.getNodes(Node.Type.Clause); // ONLY WHEN THERE IS ONLY A DEFINITION OF A FUNCTION, NOT FOR POLIMORPHIC CALLS
		final boolean thisAnonymousRoutine = moduleName == null && routineName != null;
		final boolean allRoutines = moduleName == null && routineName == null;

		for (Node clause : clauses)
		{
			final Node routine = edg.getParent(clause);
			if (thisAnonymousRoutine && routine != name)
				continue;
			
			if (!thisAnonymousRoutine && !allRoutines)
			{
				final Node module = edg.getParent(routine);
				final Node.Type moduleType = module.getType();
				if (moduleType != Node.Type.Module)
					continue;
				final String moduleText = module.getName();
				if (!moduleName.equals(moduleText) && !moduleName.equals("_"))
					continue;
				final String routineText = routine.getName();
				if (!routineName.equals(routineText) && !routineName.equals("_"))
					continue;
			}

			final Node parameters = edg.getChild(clause, 0);
			final List<Node> parameterNodes = edg.getChildren(parameters);
			if (arguments.size() != parameterNodes.size())
				continue;

			possibleClauses.add(clause);
		}

		return possibleClauses;
	}
	private List<Node> getMatchingClauses(List<Node> possibleClauses, Node call)
	{
		final List<Node> matchingClauses = new LinkedList<Node>();

		for (Node possibleClause : possibleClauses)
			if (this.matchClause(possibleClause, call))
				matchingClauses.add(possibleClause);

		return matchingClauses;
	}
	private boolean matchClause(Node possibleClause, Node call)
	{
		final Node parameters = edg.getChild(possibleClause, 0);
		final List<Node> parameterNodes = edg.getChildren(parameters);
		final Node arguments = edg.getChild(call, 1);
		final List<Node> argumentNodes = edg.getChildren(arguments);
		if (argumentNodes.size() != parameterNodes.size())
			return false;

		for (int parameterIndex = 0; parameterIndex < parameterNodes.size(); parameterIndex++)
		{
			final Node parameter = parameterNodes.get(parameterIndex);
			final Node argument = argumentNodes.get(parameterIndex);
			final List<Node[]> matches = this.getMatches(parameter, argument);

			if (matches.isEmpty())
				return false;
		}
		return true;
	}

	/************************************/
	/*********** Output edges ***********/
	/************************************/
	private void generateOutputEdges(Node call)
	{
		final Node callResult = edg.getResFromNode(call);
		final Node callee = edg.getChild(call, Node.Type.Callee);
		final Node calleeResult = edg.getResFromNode(callee);
		final List<Node> callingFunctions = edg.getInputs(calleeResult, Direction.Forwards);

		for (Node callingFunction : callingFunctions)
		{
			final Node result = edg.getResFromNode(callingFunction);

			if (result != null)
			{
				final String routineName = edg.getAncestor(callingFunction,Node.Type.Routine).getName();
				if (routineName.equals("<constructor>"))
				{
					final GlobalVariableConstraint addConstraint = new GlobalVariableConstraint(SeekingConstraint.Operation.Add, "*");
					this.edg.addEdge(edg.getChild(callingFunction, Node.Type.ParameterOut), callResult, new Edge(Edge.Type.Output, addConstraint));
				}
				this.edg.addEdge(result, callResult, new Edge(Edge.Type.Output, new PhaseConstraint(Phase.Output)));
			}
		}
	}
	
	private List<Node> getAllClauses(ClassInfo moduleInfo, String routineName)
	{
		final List<Node> methodClauses = new LinkedList<Node>();
		final List<Node> classClauses = moduleInfo.getMethods().get(routineName);
		if (classClauses != null)
			methodClauses.addAll(classClauses);
		
		final List<ClassInfo> childrenClassInfo = moduleInfo.getChildrenClasses();
		if (childrenClassInfo.isEmpty())
			return methodClauses;
		else
		{
			final List<Node> childrenMethodClauses = new LinkedList<Node>();
			for (ClassInfo childClassInfo : childrenClassInfo)
				childrenMethodClauses.addAll(this.getAllClauses(childClassInfo, routineName));
			
			for (Node childrenMethodClause : childrenMethodClauses)
				if (!methodClauses.contains(childrenMethodClause))
					methodClauses.add(childrenMethodClause);
		}
		return methodClauses;
	}
	private List<Node> getClassClauses(ClassInfo moduleInfo, String routineName, String parentRoutineName)
	{
		final List<Node> methodClauses = new LinkedList<Node>();
		final List<Node> classClauses = moduleInfo.getMethods().get(routineName);
		
		if (classClauses != null)
			methodClauses.addAll(classClauses);
		
		final List<ClassInfo> childrenClassInfo = moduleInfo.getChildrenClasses();
		if (childrenClassInfo.isEmpty())
			return methodClauses;
		else
		{	
			List<Node> parentMethodClauses = moduleInfo.getMethods().get(parentRoutineName); 
			
			final List<Node> childrenMethodClauses = new LinkedList<Node>();
			for (ClassInfo childClassInfo : childrenClassInfo)
			{
				List<Node> childParentMethodClauses = childClassInfo.getMethods().get(parentRoutineName);
				if (childParentMethodClauses == parentMethodClauses)
					childrenMethodClauses.addAll(this.getClassClauses(childClassInfo, routineName, parentRoutineName));
			}
			
			for (Node childrenMethodClause : childrenMethodClauses)
				if (!methodClauses.contains(childrenMethodClause))
					methodClauses.add(childrenMethodClause);
		}
		return methodClauses;
	}

	/*********************************************/
	/** ERLANG INTERPROCEDURAL (NO INHERITANCE) **/
	/*********************************************/

	public void generateErlang() {
		final List<Node> calls = edg.getNodes(Node.Type.Call);

		for (Node call : calls)
		{
			this.generateInputEdgesErlang(call);
			this.generateOutputEdgesErlang(call);
		}
	}

	/************************************/
	/************ Input edges ***********/
	/************************************/
	private void generateInputEdgesErlang(Node call)
	{
		final List<Node> possibleClauses = this.getPossibleClausesErlang(call);
		Collections.sort(possibleClauses, (o1, o2) -> {
			if (o1.getId() < o2.getId())
				return -1;
			if (o1.getId() > o2.getId())
				return 1;
			return 0;
		});
		final List<Node> matchingClauses = this.getMatchingClausesErlang(possibleClauses, call);
		final Node callee = edg.getChild(call, Node.Type.Callee);
		final Node calleeResult = edg.getResFromNode(callee);
		final Node arguments = edg.getChild(call, Node.Type.Arguments);
		final List<Node> argumentNodes = edg.getChildren(arguments);
		argumentNodes.removeIf(n -> n.getType() == Node.Type.Result);

		for (Node matchingClause : matchingClauses)
		{
			final Node parameters = edg.getChild(matchingClause, Node.Type.Parameters);
			final List<Node> parameterNodes = edg.getChildren(parameters);
			parameterNodes.removeIf(n -> n.getType() == Node.Type.Result);

			this.edg.addEdge(calleeResult, matchingClause,  new Edge(Edge.Type.Input, new PhaseConstraint(Phase.Input)));
			for (int argumentIndex = 0; argumentIndex < argumentNodes.size(); argumentIndex++)
			{
				final Node argument = argumentNodes.get(argumentIndex);
				final Node parameter = parameterNodes.get(argumentIndex);
				final Node argumentResult = edg.getResFromNode(argument);
				final Node parameterResult = edg.getResFromNode(parameter);

				if (argumentResult != null && parameterResult != null)
					this.edg.addEdge(argumentResult, parameterResult, new Edge(Edge.Type.Input, new PhaseConstraint(Phase.Input)));
			}
			this.edg.addEdge(calleeResult, matchingClause, new Edge(Edge.Type.Call, new PhaseConstraint(Phase.Input)));
		}
	}
	private List<Node> getPossibleClausesErlang(Node call)
	{
		final Node callee = edg.getChild(call, Node.Type.Callee);
		final Node scopeNode = edg.getChild(callee, Node.Type.Scope);
		final Node nameNode = edg.getChild(callee, Node.Type.Name);
		final Node scopeValue = edg.getChild(scopeNode, Node.Type.Value);
		final Node routineArguments = edg.getChild(call, Node.Type.Arguments);
		final List<Node> arguments = edg.getChildrenNonResult(routineArguments);

		// Module
		final Node moduleRef = scopeValue == null ? edg.getAncestor(call, Node.Type.Module) : scopeValue;
		final Node.Type moduleRefType = moduleRef.getType();
		String moduleName = moduleRef.getName();

		// Function
		final Node name = edg.getChild(nameNode, Node.Type.Value);
		final Node.Type nameType = name.getType();
		final String routineName = name.getName();

		// - M:R()	all routines of all modules				=> ANY:ANY
		// - m:R()	all routines of module m				=> m:ANY
		// - M:r()	routine r of all modules				=> ANY:r
		// - m:r()	routine f of module m					=> module:r
		// - r()	routine f of current module				=> current:r
		// - ar()	this anonymous routine					=> current:ar 	Ex: fun(X) -> X+1 end(5)
		// - X()	all routines (including anonymous ones)	=> current:ANY

		final List<Node> possibleClauses = new LinkedList<>();

		// MODULE MUST BE A LITERAL OR IT IS THE MODULE THAT CONTAINS THE CALL
		if (moduleRefType != Node.Type.Literal && moduleRefType != Node.Type.Module)
			moduleName = "_";
		// FUNCTION NAME CAN BE (1) A LITERAL OR (2) A VARIABLE
		//		(1) A FUNCTION CALL OF A PARTICULAR FUNCTION REPRESENTED BY ITS NAME
		//		(2) A FUNCTION CALL TO:
		//			(2.1) AN ANONYMOUS NAMED FUNCTION
		//			(2.2) A CALL TO AN AD-HOC DEFINED ANONYMOUS FUNCTION
		//			(2.3) A FUNCTION STORED IN A VARIABLE (DEPENDENCE ESTABLISHED BY FLOW)

		final List<Node> routines = new LinkedList<>();
		if (nameType == Node.Type.Literal) { // (1) A FUNCTION CALL OF A PARTICULAR FUNCTION
			routines.addAll(edg.getNodes(Node.Type.Routine));
		}
		else if (nameType == Node.Type.Variable) { // (2.1) NAMED ANONYMOUS FUNCTIONS
			routines.addAll(edg.getNodes(Node.Type.AnonymousRoutine));
		}

		// CONNECT TO MATCHING ROUTINE DEFINITIONS
		// NULL NAME ROUTINES ARE ANONYMOUS NON-REFERENCED FUNCTIONS
		routines.removeIf(n -> n.getName() == null || !n.getName().equals(routineName));
		for (Node routine : routines) {
			Node module = edg.getAncestor(routine,Node.Type.Module);
			if (moduleName.equals("_") || moduleName.equals(module.getName())) {
				final List<Node> clauses = edg.getChildrenNonResult(routine);
				final Node clauseParameters = edg.getChild(clauses.get(0), Node.Type.Parameters);
				final List<Node> parameters = edg.getChildrenNonResult(clauseParameters);

				if (arguments.size() != parameters.size())
					continue;
				possibleClauses.addAll(clauses);
			}
		}

		if (possibleClauses.isEmpty()) {
			if (nameType == Node.Type.AnonymousRoutine) // (2.2) A CALL TO AN AD-HOC DEFINED ANONYMOUS FUNCTION
				possibleClauses.addAll(edg.getChildrenNonResult(name));
//			else { // (2.3) CONNECT TO ALL POTENTIAL FUNCTION DEFINITIONS
//				final List<Node> clauses = edg.getNodes(Node.Type.Clause);
//				for (Node clause : clauses){
//					Node clauseModule = edg.getAncestor(clause, Node.Type.Module);
//					if (moduleName.equals("_") || moduleName.equals(clauseModule.getName())) {
//						final Node clauseParameters = edg.getChild(clause, Node.Type.Parameters);
//						final List<Node> parameters = edg.getChildrenNonResult(clauseParameters);
//						if (arguments.size() != parameters.size())
//							continue;
//
//						possibleClauses.add(clause);
//					}
//				}
//			}
		}

		return possibleClauses;
	}
	private List<Node> getMatchingClausesErlang(List<Node> possibleClauses, Node call)
	{
		final List<Node> matchingClauses = new LinkedList<>();
		final Set<Node> ignoredRoutines = new HashSet<>();
		for (Node possibleClause : possibleClauses) {
			final Node routineNode = edg.getParent(possibleClause);
			if (ignoredRoutines.contains(routineNode))
				continue;
			MatchStatus matchingClause = this.matchClauseErlang(possibleClause, call);
			if (matchingClause.isMatch())
				matchingClauses.add(possibleClause);
			if (matchingClause.isCompleteMatch())
				ignoredRoutines.add(routineNode);
		}
		return matchingClauses;
	}
	private MatchStatus matchClauseErlang(Node possibleClause, Node call)
	{
		final Node parameters = edg.getChild(possibleClause, Node.Type.Parameters);
		final List<Node> parameterNodes = edg.getChildren(parameters);
		parameterNodes.removeIf(n -> n.getType() == Node.Type.Result);

		final Node arguments = edg.getChild(call, Node.Type.Arguments);
		final List<Node> argumentNodes = edg.getChildren(arguments);
		argumentNodes.removeIf(n -> n.getType() == Node.Type.Result);

		if (argumentNodes.size() != parameterNodes.size())
			return new MatchStatus();

		final Map<String,String> patternMap = new HashMap<>();
		boolean allMatch = true;
		boolean allComplete = true;

		for (int parameterIndex = 0; parameterIndex < parameterNodes.size() && allMatch; parameterIndex++)
		{
			final Node parameter = parameterNodes.get(parameterIndex);
			final Node argument = argumentNodes.get(parameterIndex);

			MatchStatus matchStatus = this.isMatch(parameter, argument, patternMap);
			allMatch = allMatch && matchStatus.isMatch();
			allComplete = allComplete && matchStatus.isCompleteMatch();
		}
		if (allComplete){
			Node guard = edg.getChild(possibleClause, Node.Type.Guard);
			if (!edg.getChildren(guard).isEmpty())
				allComplete = false;
		}

		return new MatchStatus(allMatch,allComplete);
	}

	/************************************/
	/*********** Output edges ***********/
	/************************************/
	private void generateOutputEdgesErlang(Node call)
	{
		final Node callResult = edg.getResFromNode(call);
		final Node callee = edg.getChild(call, Node.Type.Callee);
		final Node calleeResult = edg.getResFromNode(callee);
		final List<Node> callingFunctions = edg.getInputs(calleeResult, Direction.Forwards);

		for (Node callingFunction : callingFunctions)
		{
			final Node result = edg.getResFromNode(callingFunction);

			if (result != null)
				this.edg.addEdge(result, callResult,  new Edge(Edge.Type.Output, new PhaseConstraint(Phase.Output)));
		}
	}
}