package eknife.erlang;

import edg.edge.Generator;
import edg.graph.*;
import edg.visitor.VoidVisitor;

import java.util.*;

public class ControlFlowGenerator extends VoidVisitor<Void> implements Generator {
	protected static final int INFINITY = -1;

	protected final Set<Node> hangingNodes = new HashSet<>();
	protected final Deque<Set<Node>> returnSet = new LinkedList<>();

	protected final Deque<Set<Node>> clauseHangingStack = new LinkedList<>();
	protected final Deque<Set<Node>> guardHangingStack = new LinkedList<>();
	protected final Deque<Set<Node>> switchStack = new LinkedList<>();
	protected final Deque<Set<Node>> switchSelectorStack = new LinkedList<>();
	protected final Deque<Boolean> switchHasDefaultStack = new LinkedList<>();

	protected boolean absorbentPattern = false;
	protected boolean absorbentCondition = false;
	protected boolean stopCaseCFG = false;
	// protected boolean ifContext = false;

	protected int ttl = INFINITY;
	protected boolean generating = false;
	protected Deque<Node> routineHanging = new LinkedList<>();
	protected Deque<Node> anonymousRoutineHanging = new LinkedList<>();

	protected final Deque<Node> lcPrevRestriction = new LinkedList<>();

	public ControlFlowGenerator(LAST graph)
	{
		super(graph);
	}

	@Override
	public void generate()
	{
		graph.getRootNode().accept(this, null);
	}

	protected void connectTo(Node n)
	{
		if (!generating)
			return;
		if (ttl > 0 || ttl == INFINITY)
		{
			makeConnections(n);
			if (ttl > 0) ttl--;
		}
		clearHanging();
		hangingNodes.add(n);
	}

	protected void makeConnections(Node n)
	{
		for (Node src : hangingNodes)
			graph.addEdge(src, n, Edge.Type.ControlFlow);
	}

	protected void clearHanging() {
		hangingNodes.clear();
	}

	@Override
	public void visitRoutine(Node n, Void arg)
	{
		assert hangingNodes.isEmpty();
		clauseHangingStack.push(new HashSet<>());
		guardHangingStack.push(new HashSet<>());
		generating = true;
		hangingNodes.add(n);
		routineHanging.push(n);
		super.visitRoutine(n, arg);
		routineHanging.pop();
		clearHanging();
		hangingNodes.addAll(clauseHangingStack.peek());
		connectTo(n);
		clauseHangingStack.pop();
		guardHangingStack.pop();
		generating = false;
		clearHanging();
	}

	@Override
	public void visitAnonymousRoutine(Node n, Void arg)
	{
		clauseHangingStack.push(new HashSet<>());
		guardHangingStack.push(new HashSet<>());
		anonymousRoutineHanging.push(n);
		connectTo(n);
		clearHanging();
		super.visitAnonymousRoutine(n, arg);
		anonymousRoutineHanging.pop();
		clearHanging();
		hangingNodes.addAll(clauseHangingStack.peek());
		connectTo(n);
		guardHangingStack.pop();
		clauseHangingStack.pop();
	}

	@Override
	public void visitClause(Node n, Void arg)
	{
		if (graph.getParent(n).equals(routineHanging.peek())) {
			clearHanging();
			hangingNodes.add(routineHanging.peek());
		}

		connectTo(n);
		clauseHangingStack.peek().addAll(hangingNodes);

		//assert returnSet.isEmpty();
		returnSet.push(new HashSet<>());
		graph.getChild(n, Node.Type.ParameterIn).accept(this, arg);
		graph.getChild(n, Node.Type.Parameters).accept(this, arg);
		Node guardNode = graph.getChild(n, Node.Type.Guard);
		guardNode.accept(this, arg);
		if (!graph.getChildren(guardNode).isEmpty())
			guardHangingStack.peek().addAll(hangingNodes);
		graph.getChild(n, Node.Type.Body).accept(this, arg);
		hangingNodes.addAll(returnSet.pop());
		graph.getChild(n, Node.Type.ParameterOut).accept(this, arg);
		hangingNodes.addAll(guardHangingStack.peek());
		guardHangingStack.peek().clear();
		connectTo(n);
	}

	@Override
	public void visitBlock(Node n, Void arg) {
		returnSet.push(new HashSet<>());
		super.visitBlock(n, arg);
		hangingNodes.addAll(returnSet.pop());
	}

	@Override
	public void visitEnclosed(Node n, Void arg)
	{
		super.visitEnclosed(n, arg);
		connectTo(n);
	}

	@Override
	public void visitEquality(Node n, Void arg)
	{
		new EDG(graph).getChild(n, Node.Type.Value).accept(this, arg);
		new EDG(graph).getChild(n, Node.Type.Pattern).accept(this, arg);
		connectTo(n);
	}

	@Override
	public void visitOperation(Node n, Void arg)
	{
		super.visitOperation(n, arg);
		connectTo(n);
	}

	@Override
	public void visitDataConstructor(Node n, Void arg)
	{
		super.visitDataConstructor(n, arg);
		connectTo(n);
		if (graph.isPatternZone(n) && !this.isMatchExprPattern(n))
			guardHangingStack.peek().addAll(hangingNodes);
	}

	@Override
	public void visitDataConstructorAccess(Node n, Void arg)
	{
		super.visitDataConstructorAccess(n, arg);
		connectTo(n);
	}

	@Override
	public void visitFieldAccess(Node n, Void arg)
	{
		new EDG(graph).getChild(n, Node.Type.Index).accept(this, arg);
		new EDG(graph).getChild(n, Node.Type.Variable).accept(this, arg);
	}

	@Override
	public void visitSwitch(Node n, Void arg)
	{
		returnSet.push(new HashSet<>());
		switchStack.push(new HashSet<>());
		graph.getChild(n, Node.Type.Selector).accept(this, arg);
		graph.getChild(n, Node.Type.Cases).accept(this, arg);
		clearHanging();
		hangingNodes.addAll(returnSet.pop());
		connectTo(n);
	}

	@Override
	public void visitCases(Node n, Void arg){
		absorbentPattern = false;
		absorbentCondition = false;
		stopCaseCFG = false;

		guardHangingStack.push(new HashSet<>());
		super.visitCases(n, arg);
		returnSet.peek().addAll(Set.copyOf(guardHangingStack.pop()));

		stopCaseCFG = false;
	}

	@Override
	public void visitCase(Node n, Void arg)
	{
		if(!stopCaseCFG) {
			hangingNodes.addAll(guardHangingStack.peek());
			guardHangingStack.peek().clear();
			graph.getChild(n, Node.Type.Selectable).accept(this, arg);
			graph.getChild(n, Node.Type.Guard).accept(this, arg);
			guardHangingStack.peek().addAll(Set.copyOf(hangingNodes));

			// SAVE THE ABSORBENT CASE STATE AND RESTORE IT AFTER BODY
			boolean prevAbsorbentPattern = absorbentPattern;
			boolean prevAbsorbentCondition = absorbentCondition;
			graph.getChild(n, Node.Type.Body).accept(this, arg);
			absorbentCondition = prevAbsorbentCondition;
			absorbentPattern = prevAbsorbentPattern;
			stopCaseCFG = absorbentPattern && absorbentCondition;

			absorbentPattern = false;
			absorbentCondition = false;

			returnSet.peek().addAll(Set.copyOf(hangingNodes));
		}
	}

	@Override
	public void visitSelectable(Node n, Void arg) {
		if (graph.getChildren(n).isEmpty() || this.isAbsorbentPattern(graph.getChild(n, 0)))
			absorbentPattern = true;
		else
			absorbentPattern = false;
		super.visitSelectable(n, arg);
	}

	@Override
	public void visitGuard(Node n, Void arg){
		if (this.isAbsorbentCondition(n))
			absorbentCondition = true;
		super.visitGuard(n,arg);
	}

	@Override
	public void visitCall(Node n, Void arg)
	{
		graph.getChild(n, Node.Type.Callee).accept(this, arg);
		graph.getChild(n, Node.Type.ArgumentIn).accept(this, arg);
		graph.getChild(n, Node.Type.Arguments).accept(this, arg);
		graph.getChild(n, Node.Type.ArgumentOut).accept(this, arg);
		connectTo(n);
	}

	@Override
	public void visitCallee(Node n, Void arg)
	{
		super.visitCallee(n, arg);
		connectTo(n);
	}

	@Override
	public void visitVariable(Node n, Void arg)
	{
		connectTo(n);
		if (graph.isPatternZone(n) && !this.isMatchExprPattern(n) &&
				((Variable) n).getContext() == Variable.Context.Use)
				guardHangingStack.peek().addAll(hangingNodes);
	}

	@Override
	public void visitLiteral(Node n, Void arg)
	{
		connectTo(n);
		if (graph.isPatternZone(n) && !this.isMatchExprPattern(n))
			guardHangingStack.peek().addAll(hangingNodes);
	}

	@Override
	public void visitReturn(Node n, Void arg)
	{
		super.visitReturn(n, arg);
		connectTo(n);
		returnSet.peek().add(n);
		clearHanging();
	}

	@Override
	public void visitParameterIn(Node n, Void arg)
	{
		connectTo(n);
	}

	@Override
	public void visitParameterOut(Node n, Void arg)
	{
		connectTo(n);
	}

	@Override
	public void visitArgumentIn(Node n, Void arg)
	{
		connectTo(n);
	}

	@Override
	public void visitArgumentOut(Node n, Void arg)
	{
		connectTo(n);
	}


	// LIST COMPREHENSIONS AND LISTS IN LIST COMPREHENSIONS
	@Override
	public void visitListComprehension(Node n, Void arg) {
		lcPrevRestriction.push(null);
		graph.getChild(n, Node.Type.Restrictions).accept(this, arg);
		graph.getChild(n, Node.Type.Value).accept(this, arg);
		connectTo(lcPrevRestriction.pop());

		Node restrictions = graph.getChild(n, Node.Type.Restrictions);
		Node firstGenerator = graph.getChild(restrictions, Node.Type.Generator);
		//connectTo(graph.getChild(firstGenerator, Node.Type.Iterator));

		clearHanging();
		hangingNodes.add(firstGenerator);
		connectTo(n);
	}
	@Override
	public void visitRestrictions(Node n, Void arg)
	{
		graph.getChildren(n, Node.Type.Generator).stream()
				.forEach(node -> node.accept(this, arg));
		graph.getChildren(n, Node.Type.Filter).stream()
				.forEach(node -> node.accept(this, arg));
	}

	@Override
	public void visitFilter(Node n, Void arg) {
		super.visitFilter(n,arg);
		connectTo(lcPrevRestriction.pop());
		clearHanging();
		Node filterExpression = graph.getChild(n, Node.Type.Value);
		hangingNodes.add(filterExpression);
		lcPrevRestriction.push(filterExpression);
	}

	@Override
	public void visitGenerator(Node n, Void arg)
	{
		boolean isFirstGenerator = lcPrevRestriction.peek() == null;

		new EDG(graph).getChild(n, Node.Type.Iterator).accept(this, arg);
		new EDG(graph).getChild(n, Node.Type.Variable).accept(this, arg);
		connectTo(n);

		if (!isFirstGenerator)
			connectTo(lcPrevRestriction.pop());
		clearHanging();
		hangingNodes.add(n);
		lcPrevRestriction.push(n);
	}
	@Override
	public void visitList(Node n, Void arg) {
		super.visitList(n, arg);
		connectTo(n);

		if (graph.isPatternZone(n) && !this.isMatchExprPattern(n)) {
			if ((graph.getParent(n).getType() != Node.Type.List) &&
				graph.getChildren(n).isEmpty());
			guardHangingStack.peek().addAll(hangingNodes);
		}
	}

	private boolean isAbsorbentPattern(Node n) {
		if (n.getType() == Node.Type.Variable &&
				((Variable) n).getContext() == Variable.Context.Definition)
			return true;

		return false;
	}

	private boolean isAbsorbentCondition(Node n) {
		List<Node> guardChildren = graph.getChildren(n);
		if (guardChildren.isEmpty())
			return true;

		Node exprNode = graph.getChildren(
							graph.getChildren(
									guardChildren.get(0)
							).get(0)
						).get(0);

		return exprNode.getType() == Node.Type.Literal &&
				exprNode.getName().equals("true");
	}

	private boolean isMatchExprPattern(Node n) {
		Node containerExpr = getContainerExpr(n);
		switch (containerExpr.getType()) {
			case Equality:
			case Generator:
				return true;
			default:
				return false;
		}
	}
	private Node getContainerExpr(Node n) {
		Node parent = graph.getParent(n);
		while (parent != null && !isContainerExpr(parent)) {
			parent = graph.getParent(parent);
		}
		if (parent == null)
			throw new IllegalStateException("Cannot exist a pattern that is not contained in another expression with patterns");
		return parent;
	}

	private boolean isContainerExpr(Node n) {
		switch (n.getType()){
			case Equality:
			case Generator:
			case Clause:
			case Switch:
			case CatchClause:
				return true;
			default:
				return false;
		}
	}
}